"""Calibration analysis scripts.

This submodule is used to extract basic calibration constants, such as:
- Track dE/dx profile
- Wire transparency map
- MCS angular resolution calibration
- ...
"""

from .mcs import *
