"""Contains reusable layers for various models. Layers are classified into three folders:

* `common` reusable by any models
* `cnn` reusable by CNN-based models
* `cluster` reusable by CNN-based clustering models
* `gnn` reusable by GNN-based models
"""
